/**
 * TP1.java
 * 
 * Time-stamp: <2012-12-20 23:29:16 ftaiani>
 * @author Francois Taiani   <f.taiani@lancaster.ac.uk>
 * $Id$
 * 
 */

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

class ArrayListWrapper<T> {
  private List<T> internalList = new ArrayList<T>();
  private Random rand = new Random();

  synchronized public void addRandom(T l) {
    internalList.add(rand.nextInt(internalList.size()+1),l);
    // we wake up all possibly waiting threads
    // Exercise: What do we need to wake all threads, and not only one?
    if (internalList.size()==1) notifyAll();
  } // EndMethod addRandom

  synchronized public void removeRandom() throws InterruptedException {
    while(internalList.isEmpty()) wait();
    internalList.remove(rand.nextInt(internalList.size()));
  } // EndMethod addRandom
  
  synchronized public int size() {
    return internalList.size(); // a case of delegation
  }
  
} // EndClass ArrayListWrapper

class Inserter extends Thread {
  public void run() {
    for(int i=0;i<10000;i++) {
      // insert the thread's ID at a random position in the list
      TP1.globalList.addRandom(this.getId());
    }
	}
} // EndClass Inserter

class Remover extends Thread {
  public void run() {
    try {
      for(int i=0;i<10000;i++) {
        // remove an ID at a random position in the list
        TP1.globalList.removeRandom();
      }
    }
    catch(InterruptedException e) {
      e.printStackTrace(); // thread terminates if interrupted
    }
  }
} // EndClass Remover


class Observer extends Thread {

  // Here 'volatile' means that continueToWork will be accessed /
  // modified by several threads and should therefore behave as an
  // atomic concurrent object. (Not guaranteed without volatile, in
  // particular because of caching by threads.)

  public volatile boolean continueToWork = true ;

  public void run() {
    try {
      while(continueToWork) {
        System.err.println("Size of list: " + TP1.globalList.size());
        Thread.sleep(200);
      }
    } 
    catch(InterruptedException e) {
      e.printStackTrace(); // thread terminates if interrupted
    }
  }
}

/**
 * @author Francois Taiani   <f.taiani@lancaster.ac.uk>
 */
public class TP1 {

  static public ArrayListWrapper<Long> globalList = new ArrayListWrapper<Long>();
  
  public static void main(String[] args) throws InterruptedException {
    Observer observe = new Observer();
    observe.start();

    List<Thread> allThreads = new ArrayList<Thread>();
    
    // Creating the 10 Inserters, 10 Removers
    for(int i=0;i<10;i++) {
      Thread insert = new Inserter();
      Thread remove = new Remover();
      remove.start();
      insert.start();
      // keeping a reference on the threads we have created
      allThreads.add(insert);
      allThreads.add(remove);
    } // EndFor

    // waiting for all threads to terminate
    for(Thread t : allThreads) { t.join(); }

    Thread.sleep(800); // so we can see that the list has stabilised
    observe.continueToWork = false ; // telling observer thread to stop

  } // EndMain
  
} // EndClass TP1
